##############################################################################
#
# Copyright (c) 2015 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import re
import base64
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
import collections
import logging
import optparse
import os.path
import pkg_resources
import subprocess
import sys
try:
    # py2
    from urlparse import urlparse
    from httplib import HTTPConnection
    from httplib import HTTPSConnection
except ImportError:
    # py3
    from urllib.parse import urlparse
    from http.client import HTTPConnection
    from http.client import HTTPSConnection
try:
    import pkg_resources._vendor.packaging.version
    packaging = pkg_resources._vendor.packaging
except ImportError:
    # fallback to naturally-installed version; allows system packagers to
    #  omit vendored packages.
    import packaging.version

logger = logging.Logger('build')
formatter = logging.Formatter('%(levelname)s - %(message)s')

BUILD_SECTION = 'build'


def getBasicAuthString(username, password):
    s = '%s:%s' % (username, password)
    bs = base64.b64encode(s.encode('utf-8'))
    return bs.decode('utf-8')


def parseVersion(s):
    """pkg_resources.parse_version is deprecated"""
    component_re = re.compile(r'(\d+ | [a-z]+ | \.| -)', re.VERBOSE)
    replace = {
        'pre': 'c',
        'preview': 'c',
        '-': 'final-',
        'rc': 'c',
        'dev': '@',
    }.get

    def _parse_version_parts(s):
        for part in component_re.split(s):
            part = replace(part, part)
            if not part or part == '.':
                continue
            if part[:1] in '0123456789':
                # pad for numeric comparison
                yield part.zfill(8)
            else:
                yield '*'+part

        # ensure that alpha/beta/candidate are before final
        yield '*final'

    def old_parse_version(s):
        parts = []
        for part in _parse_version_parts(s.lower()):
            if part.startswith('*'):
                # remove '-' before a prerelease tag
                if part < '*final':
                    while parts and parts[-1] == '*final-':
                        parts.pop()
                # remove trailing zeros from each series of numeric parts
                while parts and parts[-1] == '00000000':
                    parts.pop()
            parts.append(part)
        return tuple(parts)

    for part in old_parse_version(str(s)):
        yield part


def getFilePath(fName, options):
    """Apply storage path if given and not applied"""
    if options.storagePath and not fName.startswith(options.storagePath):
        fName = '%s/%s' % (options.storagePath, fName)
    return fName


def do(cmd, cwd=None, captureOutput=True):
    logger.debug('Command: ' + cmd)
    if captureOutput:
        stdout = stderr = subprocess.PIPE
    else:
        stdout = stderr = None
    p = subprocess.Popen(
        cmd, stdout=stdout, stderr=stderr,
        shell=True, cwd=cwd)
    stdout, stderr = p.communicate()
    if stdout is None:
        stdout = "See output above"
    if stderr is None:
        stderr = "See output above"
    if p.returncode != 0:
        logger.error('An error occurred while running command: %s' %cmd)
        logger.error('Error Output: \n%s' % stderr)
        sys.exit(p.returncode)
    logger.debug('Output: \n%s' % stdout)
    return stdout


class SVN(object):

    user = None
    passwd = None
    forceAuth = False

    #TODO: spaces in urls+folder names???

    def __init__(self, user=None, passwd=None, forceAuth=False,
        trustServerCert=False):
        self.user = user
        self.passwd = passwd
        self.forceAuth = forceAuth
        self.trustServerCert = trustServerCert

    def getCommand(self, command, options):
        """Replace trust server cert marker"""
        command += ' --non-interactive'
        # add authentication
        if self.user:
            command += ' --username %s --password %s' % (self.user, self.passwd)
            if self.forceAuth:
                command += ' --no-auth-cache'
        # add trusted server certificate
        if self.trustServerCert:
            command += ' --trust-server-cert'
        if options:
            command += ' %s' % options
        return command

    def info(self, url):
        command = 'svn info'
        options = '--xml %s' % url
        command = self.getCommand(command, options)
        return do(command)

    def ls(self, url):
        command = 'svn ls'
        options = '--xml %s' % url
        command = self.getCommand(command, options)
        return do(command)

    def cp(self, fromurl, tourl, comment):
        command = 'svn cp'
        options = '-m "%s" %s %s' % (comment, fromurl, tourl)
        command = self.getCommand(command, options)
        do(command)

    def co(self, url, folder):
        command = 'svn co'
        options = '%s %s' % (url, folder)
        command = self.getCommand(command, options)
        do(command)

    def ci(self, folder, comment):
        command = 'svn ci'
        options = '-m "%s" %s' % (comment, folder)
        command = self.getCommand(command, options)
        do(command)


# Python 2/3 compatible input
try:
    string_input = raw_input  # Python 2
except NameError:
    string_input = input  # Python 3


def getInput(prompt, default, useDefaults):
    if useDefaults:
        return default
    defaultStr = ''
    if default:
        defaultStr = ' [' + default + ']'
    value = string_input(prompt + defaultStr + ': ')
    if not value:
        return default
    return value


def uploadContent(content, filename, url, username, password,
                  offline, method, headers=None):
    if offline:
        logger.info('Offline: File `%s` not uploaded.' %filename)
        return

    logger.debug('Uploading `%s` to %s' %(filename, url))
    pieces = urlparse(url)
    Connection = HTTPConnection
    if pieces[0] == 'https':
        Connection = HTTPSConnection

    credentials = ('%s:%s' % (username, password)).encode('utf-8')
    base64string = base64.b64encode(credentials).decode('utf-8')

    if headers is None:
        headers = {}

    headers["Authorization"] = "Basic %s" % base64string

    conn = Connection(pieces[1])
    conn.request(
        method,
        pieces[2],
        content,
        headers)

    response = conn.getresponse()
    if ((response.status != 201 and method == 'PUT')
        or response.status != 200 and method == 'POST'):
        if response.status == 400 and 'KeyError' in response.reason:
            logger.info(
                "File upload failed with Code: %i (%s). Does this file "
                "already exist?" % (response.status, response.reason))
        else:
            logger.error("Error uploading file. Code: %i (%s)" % (
                response.status, response.reason))
    else:
        logger.info("File uploaded: %s" % filename)


def uploadFile(path, url, username, password, offline, method='PUT',
               headers=None):
    filename = os.path.split(path)[-1]

    uploadContent(open(path, 'r').read(),
                  filename, url+'/'+filename,
                  username, password, offline, method, headers=headers)


def guessNextVersion(version):
    #pkg_resources.parse_version(version)
    pieces = parseVersion(version)
    newPieces = []
    for piece in pieces:
        try:
            newPieces.append(int(piece))
        except ValueError:
            break
    newPieces += [0]*(3-len(newPieces))
    newPieces[-1] += 1
    newVersion = '.'.join([str(piece) for piece in newPieces])
    logger.debug('Last Version: %s -> %s' %(version, newVersion))
    return newVersion


leading_blank_lines = re.compile(r"^(\s*\n)+")
leading_blank_lines = re.compile(r"^(\s*\n)+")


class NonDestructiveRawConfigParser(configparser.RawConfigParser):
    """Config parser preserving whitespace

    TODO: empty lines or lines with only white space get still removed
    this means p01.build is still not useable for recipe which need to
    preserve empty lines.
    """

    def __init__(self):
        configparser.RawConfigParser.__init__(self,
            dict_type=collections.OrderedDict)

    def optionxform(self, optionstr):
        return optionstr

    def _read(self, fp, fpname):
        """Fix intend in continue line handling.

        The original config parser removes starting whitespace which will
        not work for install script content or any content as option value
        which needs to preserve leading whitespace.
        """
        cursect = None                        # None, or a dictionary
        optname = None
        lineno = 0
        e = None                              # None, or an exception
        while True:
            line = fp.readline()
            if not line:
                break # EOF

            lineno = lineno + 1

            # conditions
            if line[0] in '#;':
                # skip comment
                continue
            if line.strip() == '':
                # skip empty blank line
                continue
            if line.split(None, 1)[0].lower() == 'rem' and line[0] in "rR":
                continue

            # start collecting
            if line[0].isspace() and cursect is not None and optname:
                # continuation line
                line = line.rstrip()
                cursect[optname].append(line)
            else:
                # section header or option header?
                mo = self.SECTCRE.match(line)
                if mo:
                    sectname = mo.group('header')
                    if sectname in self._sections:
                        cursect = self._sections[sectname]
                    elif sectname == configparser.DEFAULTSECT:
                        cursect = self._defaults
                    else:
                        cursect = self._dict()
                        cursect['__name__'] = sectname
                        self._sections[sectname] = cursect
                    # So sections can't start with a continuation line
                    optname = None
                elif cursect is None:
                    if not line.strip():
                        continue
                    # no section header in the file?
                    raise configparser.MissingSectionHeaderError(fpname,
                        lineno, line)
                else:
                    mo = self._optcre.match(line)
                    if mo:
                        optname, vi, optval = mo.group('option', 'vi', 'value')
                        optname = self.optionxform(optname.rstrip())
                        # This check is fine because the OPTCRE cannot
                        # match if it would set optval to None
                        if optval is not None:
                            if vi in ('=', ':') and ';' in optval:
                                # ';' is a comment delimiter only if it follows
                                # a spacing character
                                pos = optval.find(';')
                                if pos != -1 and optval[pos-1].isspace():
                                    optval = optval[:pos]
                            optval = optval.rstrip()
                            # allow empty values
                            if optval == '""':
                                optval = ''
                            cursect[optname] = [optval]
                    elif not (optname or line.strip()):
                        # skip blank line after section start
                        continue
                    else:
                        # a non-fatal parsing error occurred.  set up the
                        # exception but keep going. the exception will be
                        # raised at the end of the file and will contain a
                        # list of all bogus lines
                        if not e:
                            e = configparser.ParsingError(fpname)
                        e.append(lineno, repr(line))
        # if any parsing errors occurred, raise an exception
        if e:
            raise e

        # join the multi-line values collected while reading
        all_sections = [self._defaults]
        all_sections.extend(list(self._sections.values()))
        for options in all_sections:
            for name, val in list(options.items()):
                if isinstance(val, list):
                    options[name] = '\n'.join(val)

    def writeFile(self, fName, options):
        """Write an .ini-format representation of the configuration state.

        Enhanced support for buildout specfic "+=" key/value options e.g.
        [buildout]
        extends = foo-parts.cfg
        parts += foo

        Previous version generated parts + = foo. But this now fails since
        buildout >= 2.0

        """
        fName = getFilePath(fName, options)
        fp = open(fName, 'w')
        if self._defaults:
            fp.write("[%s]\n" % configparser.DEFAULTSECT)
            for (key, value) in list(self._defaults.items()):
                fp.write("%s = %s\n" % (key, str(value).replace('\n', '\n\t')))
            fp.write("\n")
        for section in self._sections:
            fp.write("[%s]\n" % section)
            for (key, value) in list(self._sections[section].items()):
                if key == "__name__":
                    continue
                if (value is not None) or (self._optcre == self.OPTCRE):
                    # NOTE: this supports parts += ... and prevents to split
                    # the += into "+ =" which would fail later using zc.buildout
                    if key.endswith(' +') or key.endswith(' -'):
                        key = "= ".join((key, str(value)))
                    else:
                        key = " = ".join((key, str(value)))
                fp.write("%s\n" % (key))
            fp.write("\n")


parser = optparse.OptionParser()
parser.add_option(
    "-c", "--config-file", action="store",
    dest="configFile", metavar="FILE",
    help="The file containing the configuration of the project.")

parser.add_option(
    "-q", "--quiet", action="store_true",
    dest="quiet", default=False,
    help="When specified, no messages are displayed.")

parser.add_option(
    "-v", "--verbose", action="store_true",
    dest="verbose", default=False,
    help="When specified, debug information is created.")

parser.add_option(
    "-d", "--use-defaults", action="store_true",
    dest="useDefaults", default=False,
    help="When specified, no user input is required and the defaults are used.")

parser.add_option(
    "-o", "--offline-mode", action="store_true",
    dest="offline", default=False,
    help="When set, no server commands are executed.")

parser.add_option(
    "-n", "--next-version", action="store_true",
    dest="nextVersion", default=False,
    help="When set, the system guesses the next version to generate.")

parser.add_option(
    "--force-version", action="store",
    dest="forceVersion", default="", metavar="VERSION",
    help="Force one common version through all packages and configs.")

parser.add_option(
    "--default-package-version", action="store",
    dest="defaultPackageVersion", default="", metavar="VERSION",
    help="Set a default package version for not yet released ones.")

parser.add_option(
    "--force-svnauth", action="store_true",
    dest="forceSvnAuth", default=False,
    help="Force svn authentication with svn-repos- credentials.")

parser.add_option(
    "-b", "--use-branch", action="store",
    dest="branch", metavar="BRANCH", default=None,
    help="When specified, this branch will be always used.")

parser.add_option(
    "-i", "--independent-branches", action="store_true",
    dest="independent", metavar="INDEPENDENT", default=False,
    help=("When specified, the system makes sure the last release is based "
         "on the given branch."))

parser.add_option(
    "--no-upload", action="store_true",
    dest="noUpload", default=False,
    help="When set, the generated configuration files are not uploaded.")

parser.add_option(
    "--no-branch-update", action="store_true",
    dest="noBranchUpdate", default=False,
    help=("When set, the branch is not updated with a new version after a "
         "release is created."))

parser.add_option(
    "-s", "--storage-path", action="store",
    dest="storagePath", default=None,
    help=("Generate config files in this folder"))



##############################################################################
#
# Copyright (c) 2015 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import bs4
import configparser
import base64
import logging
import optparse
import os
import pkg_resources
import re
import shutil
import sys
import stat
import tempfile
try:
    # py2
    from urllib2 import urlopen
    from urllib2 import Request
    from urlparse import urljoin
except ImportError:
    # py3
    from urllib.request import urlopen
    from urllib.request import Request
    from urllib.parse import urljoin
try:
    import pkg_resources._vendor.packaging.version
    packaging = pkg_resources._vendor.packaging
except ImportError:
    # fallback to naturally-installed version; allows system packagers to
    #  omit vendored packages.
    import packaging.version

from xml.etree import ElementTree
from p01.build import base

logger = base.logger

is_win32 = sys.platform == 'win32'

# WindowsError is not available in Python 3, use OSError instead
try:
    WindowsError
except NameError:
    WindowsError = OSError


def checkRO(function, path, excinfo):
    if (function == os.remove
        and excinfo[0] == WindowsError
        and getattr(excinfo[1], 'winerror', None) == 5):
        #Access is denied
        #because it's a readonly file
        os.chmod(path, stat.S_IWRITE)
        os.remove(path)


def rmtree(dirname):
    if is_win32:
        shutil.rmtree(dirname, ignore_errors=False, onerror=checkRO)
    else:
        shutil.rmtree(dirname)


class PackageBuilder(object):

    pkg = None
    customPath = None
    options = None

    uploadType = 'internal'
    uploadFormat = None
    packageIndexUrl = None
    packageIndexUsername = None
    packageIndexPassword = None

    svnRepositoryUrl = None
    trustServerCert = False

    tagLayout = 'flat'
    svn = None

    #filled by runCLI, as an info for build.py
    branchUrl = None
    branchRevision = None

    def __init__(self, pkg, options):
        self.pkg = pkg
        self.options = options

    def getTagURL(self, version):
        reposUrl = self.svnRepositoryUrl
        separator = '-'
        if self.tagLayout == 'flat':
            separator = '-'
        elif self.tagLayout == 'subfolder':
            separator = '/'
        if self.customPath:
            reposUrl = urljoin(reposUrl, self.customPath)
            tagUrl = reposUrl.split('%s')[0] + 'tags/%s%s%s' %(
                self.pkg, separator, version)
        else:
            tagUrl = reposUrl + 'tags/%s%s%s' %(self.pkg, separator, version)
        logger.debug('Tag URL: ' + tagUrl)
        return tagUrl

    def getBranchURL(self, branch):
        reposUrl = self.svnRepositoryUrl
        if self.customPath:
            reposUrl = urljoin(reposUrl, self.customPath)
            branchUrl = reposUrl %('branches/' + branch)
            if branch == 'trunk':
                branchUrl = reposUrl %branch
        else:
            branchUrl = reposUrl + 'branches/' + branch + '/' + self.pkg + '/'
            if branch == 'trunk':
                branchUrl = reposUrl + 'trunk/' + self.pkg
        logger.debug('Branch URL: ' + branchUrl)
        return branchUrl

    def getRevision(self, url):
        #xml = base.do('svn info --xml ' + url)
        xml = self.svn.info(url)
        elem = ElementTree.fromstring(xml)
        revision = elem.find("entry").find("commit").get("revision")
        if not revision:
            revision = 0
        else:
            revision = int(revision)
        logger.debug('Revision for %s: %i' %(url, revision))

        repoRevision = elem.find("entry").get("revision")
        if not repoRevision:
            repoRevision = 0
        else:
            repoRevision = int(repoRevision)
            logger.debug('Repo Revision for %s: %i' %(url, repoRevision))
        return (repoRevision, revision)

    def findVersions(self):
        if self.options.offline:
            logger.info('Offline: Skip looking for versions.')
            return []

        logger.debug('Package Index: ' + self.packageIndexUrl)
        req = Request(self.packageIndexUrl)

        if self.packageIndexUsername:
            credentials = ('%s:%s' % (self.packageIndexUsername,
                           self.packageIndexPassword)).encode('utf-8')
            base64string = base64.b64encode(credentials).decode('utf-8')
            req.add_header("Authorization", "Basic %s" % base64string)

        soup = bs4.BeautifulSoup(urlopen(req).read(), "html.parser")

        VERSION = re.compile(self.pkg+r'-(\d+\.\d+(\.\d+){0,2})')

        simplePageUrl = None

        versions = []
        for tag in soup('a'):
            cntnt = str(tag.contents[0]) # str: re does not like non-strings

            if cntnt == self.pkg:
                try:
                    simplePageUrl = tag['href']
                except KeyError:
                    pass

            m = VERSION.search(cntnt)
            if m:
                versions.append(m.group(1))

        if len(versions) == 0 and simplePageUrl:
            #we probably hit a PYPI-like simple index
            #reload the linked page, check again for versions
            req = Request(simplePageUrl)

            if self.packageIndexUsername:
                credentials = ('%s:%s' % (self.packageIndexUsername,
                               self.packageIndexPassword)).encode('utf-8')
                base64string = base64.b64encode(credentials).decode('utf-8')
                req.add_header("Authorization", "Basic %s" % base64string)

            soup = bs4.BeautifulSoup(urlopen(req).read(), "html.parser")

            for tag in soup('a'):
                cntnt = str(tag.contents[0]) # str: re does not like non-strings

                m = VERSION.search(cntnt)
                if m:
                    versions.append(m.group(1))

        logger.debug('All versions: ' + ' '.join(versions))

        # filter versions by ones that came from the branch we are building from.
        if self.options.branch and '-' in self.options.branch:
            branchVersion = self.options.branch.split('-')[-1]
            branchVersionParts = tuple(
                [p for p in base.parseVersion(branchVersion)
                 if not p.startswith('*')]
                 )
            def fromBranch(v):
                v = base.parseVersion(v)
                return v[:len(branchVersionParts)] == branchVersionParts
            versions = list(filter(fromBranch, versions))
        return sorted(versions, key=lambda x: packaging.version.Version(x))

    def getBranches(self):
        if self.options.offline:
            logger.info('Offline: Skip looking for branches.')
            return []
        url = self.svnRepositoryUrl
        if self.customPath:
            url = urljoin(url, self.customPath.split('%s')[0])
        url += 'branches'
        logger.debug('Branches URL: ' + url)

        #xml = base.do('svn ls --xml ' + url)
        xml = self.svn.ls(url)
        elem = ElementTree.fromstring(xml)

        branches = [elem.text for elem in elem.findall('./list/entry/name')]

        #if self.svnRepositoryUsername:
        #    base64string = base64.encodestring('%s:%s' % (
        #        self.svnRepositoryUsername, self.svnRepositoryPassword))[:-1]
        #    req.add_header("Authorization", "Basic %s" % base64string)
        #
        #soup = bs4.BeautifulSoup(urllib2.urlopen(req).read(), "html.parser")
        #branches  = [tag.contents[0][:-1]
        #             for tag in soup('ul')[0]('a')
        #             if tag.contents[0] != '..']
        logger.debug('Branches: ' + ' '.join(branches))
        return branches

    def hasChangedSince(self, version, branch):
        # check if svn revision gets changed on the branch after the tag
        # was created, so that the branch always has a later revision. Note
        # that our last release was updating the version in setup.py which also
        # forces a change after adding the tag. So let's check the source
        # directory instead.
        branchUrl = self.getBranchURL(branch) + '/src'
        tagUrl = self.getTagURL(version)
        changed = self.getRevision(branchUrl)[1] > self.getRevision(tagUrl)[1]
        if changed:
            logger.info(
                'Branch %r changed since the release of version %s' %(
                branch, version))
        return changed

    def isLastReleaseFromBranch(self, version, branch):
        # check if the dev marked version in setup.py from the given branch
        # compares with our version we will guess. If so, this means no
        # other branch was used for release this package.
        branchURL = self.getBranchURL(branch)
        if branchURL.endswith('/'):
            branchURL = branchURL[:-1]
        pyURL = '%s/setup.py' % branchURL
        req = Request(pyURL)
        if self.packageIndexUsername:
            credentials = ('%s:%s' % (self.packageIndexUsername,
                           self.packageIndexPassword)).encode('utf-8')
            base64string = base64.b64encode(credentials).decode('utf-8')
            req.add_header("Authorization", "Basic %s" % base64string)
        setuppy = urlopen(req).read().decode('utf-8')
        nextVersion = re.search("version ?= ?'(.*)',", setuppy)
        if not nextVersion:
            logger.error("No version =  found in setup.py, cannot update!")
            # prevent mess up, force ensure new release
            return False
        else:
            nextVersion = nextVersion.groups()[0]
            setupVersion = '%sdev' % base.guessNextVersion(version)
            if setupVersion == nextVersion:
                return True
            else:
                logger.info("Last release %s wasn't released from branch %r " % (
                    version, branch))
                return False

    def createRelease(self, version, branch):
        logger.info('Creating release %r for %r from branch %r' %(
            version, self.pkg, branch))
        # 0. Skip creating releases in offline mode.
        if self.options.offline:
            logger.info('Offline: Skip creating a release.')
            return
        # 1. Create Release Tag
        branchUrl = self.getBranchURL(branch)
        tagUrl = self.getTagURL(version)

        logger.info('Creating release tag')
        #TODO: destination folder might not exist... create it
        self.svn.cp(branchUrl, tagUrl, "Create release tag %s." % version)
        #base.do('svn cp -m "Create release tag %s." %s %s' %(
        #    version, branchUrl, tagUrl))

        # 2. Download tag
        buildDir = tempfile.mkdtemp()
        tagDir = os.path.join(buildDir, '%s-%s' %(self.pkg, version))
        self.svn.co(tagUrl, tagDir)
        #base.do('svn co %s %s' %(tagUrl, tagDir))

        # 3. Create release
        # 3.1. Remove setup.cfg
        logger.info("Updating tag version metadata")
        setupCfgPath = os.path.join(tagDir, 'setup.cfg')
        if os.path.exists(setupCfgPath):
            os.remove(setupCfgPath)
        # 3.2. Update the version
        with open(os.path.join(tagDir, 'setup.py'), 'r') as f:
            setuppy = f.read()
        setuppy = re.sub(
            "version ?= ?'(.*)',", "version = '%s'," %version, setuppy)
        with open(os.path.join(tagDir, 'setup.py'), 'w') as f:
            f.write(setuppy)
        # 3.3. Check it all in
        self.svn.ci(tagDir, "Prepare for release %s." % version)
        #base.do('svn ci -m "Prepare for release %s." %s' %(version, tagDir))

        if self.uploadFormat is not None:
            frmts = ' --formats=%s' % self.uploadFormat
        else:
            frmts = ''

        # 4. Upload the distribution
        if self.uploadType == 'internal':
            # 3.4. Create distribution
            logger.info("Creating release tarball")
            base.do('python setup.py sdist%s' % frmts, cwd = tagDir)
            # https://docs.python.org/2/distutils/sourcedist.html
            if self.uploadFormat in [
                'zip', 'gztar', 'bztar', 'ztar', 'tar']:
                if self.uploadFormat == 'zip':
                    ext = 'zip'
                elif self.uploadFormat == 'gztar':
                    ext = 'tar.gz'
                elif self.uploadFormat == 'bztar':
                    ext = 'tar.bz2'
                elif self.uploadFormat == 'ztar':
                    ext = 'tar.Z'
                elif self.uploadFormat == 'tar':
                    ext = 'tar'
            elif is_win32:
                ext = 'zip'
            else:
                ext = 'tar.gz'
            distributionFileName = os.path.join(
                tagDir, 'dist', '%s-%s.%s' %(self.pkg, version, ext))
            if not self.options.noUpload:
                logger.info("Uploading release.")
                base.uploadFile(
                    distributionFileName,
                    self.packageIndexUrl,
                    self.packageIndexUsername, self.packageIndexPassword,
                    self.options.offline)
        elif self.uploadType == 'setup.py':
            # 3.4. Create distribution and upload in one step
            logger.info("Uploading release to PyPI.")
            # definitely DO NOT register!!!
            base.do('python setup.py sdist%s upload' % frmts, cwd = tagDir)
        else:
            logger.warn('Unknown uploadType: ' + self.uploadType)

        # 5. Update the start branch to the next development (dev) version
        if not self.options.noBranchUpdate:
            logger.info("Updating branch version metadata")
            # 5.1. Check out the branch.
            branchDir = os.path.join(buildDir, 'branch')
            self.svn.co(branchUrl, branchDir)
            #base.do('svn co --non-recursive %s %s' %(branchUrl, branchDir))
            # 5.2. Get the current version.
            with open(os.path.join(branchDir, 'setup.py'), 'r') as f:
                setuppy = f.read()
            currVersion = re.search("version ?= ?'(.*)',", setuppy)
            if not currVersion:
                logger.error("No version =  found in setup.py, cannot update!")
            else:
                currVersion = currVersion.groups()[0]
                # 5.3. Update setup/py to the next version of the currently
                #      released one
                newVersion = base.guessNextVersion(version) + 'dev'
                setuppy = re.sub(
                    "version ?= ?'(.*)',", "version = '%s'," %newVersion, setuppy)
                with open(os.path.join(branchDir, 'setup.py'), 'w') as f:
                    f.write(setuppy)
                # 5.4. Check in the changes.
                self.svn.ci(branchDir, "Update version number to %s." % newVersion)
                #base.do('svn ci -m "Update version number to %s." %s' %(
                #    newVersion, branchDir))

        # 6. Cleanup
        rmtree(buildDir)

    def runCLI(self, configFile, askToCreateRelease=False, forceSvnAuth=False):
        logger.info('-' * 79)
        logger.info(self.pkg)
        logger.info('-' * 79)
        logger.info('Start releasing new version of ' + self.pkg)
        # 1. Read the configuration file.
        logger.info('Loading configuration file: ' + configFile)
        config = configparser.RawConfigParser()
        config.read(configFile)
        # 1.1. Get package index info.
        self.packageIndexUrl = config.get(
            base.BUILD_SECTION, 'package-index')
        self.packageIndexUsername = config.get(
            base.BUILD_SECTION, 'package-index-username')
        self.packageIndexPassword = config.get(
            base.BUILD_SECTION, 'package-index-password')
        # 1.2. Get svn repository info.
        self.svnRepositoryUrl = config.get(
            base.BUILD_SECTION, 'svn-repos')
        try:
            self.trustServerCert = config.get(base.BUILD_SECTION,
                'svn-trust-server-cert')
        except configparser.NoOptionError:
            self.trustServerCert = False
        if forceSvnAuth:
            svnRepositoryUsername = config.get(
                base.BUILD_SECTION, 'svn-repos-username')
            svnRepositoryPassword = config.get(
                base.BUILD_SECTION, 'svn-repos-password')
            self.svn = base.SVN(svnRepositoryUsername, svnRepositoryPassword,
                forceAuth=True, trustServerCert=self.trustServerCert)
        else:
            self.svn = base.SVN(trustServerCert=self.trustServerCert)

        try:
            self.uploadType = config.get(
                base.BUILD_SECTION, 'upload-type')
        except configparser.NoOptionError:
            self.uploadType = 'internal'

        try:
            self.uploadFormat = config.get(
                base.BUILD_SECTION, 'upload-format')
        except configparser.NoOptionError:
            self.uploadFormat = None

        try:
            self.tagLayout = config.get(
                base.BUILD_SECTION, 'tag-layout')
        except configparser.NoOptionError:
            self.tagLayout = 'flat'

        # 1.3. Determine the possibly custom path.
        for pkg in config.get(base.BUILD_SECTION, 'packages').split():
            if pkg.startswith(self.pkg):
                if ':' in pkg:
                    self.customPath = pkg.split(':')[1]
                break

        # 2. Find all versions.
        versions = self.findVersions()
        logger.info('Existing %s versions: %s' % (
            self.pkg, ' | '.join(reversed(versions))))

        # 3. Determine the default version to suggest.
        defaultVersion = None

        # 3.1 Set default version based on forceVersion
        forceVersion = self.options.forceVersion
        if forceVersion:
            if forceVersion in versions:
                logger.error('Forced version %s already exists' % forceVersion)
            else:
                defaultVersion = forceVersion

        if versions and not defaultVersion:
            if self.options.nextVersion:
                # 3.2. If the branch was specified, check whether it changed
                # since the last release or if independent is set, if the last
                # release is based on the current branch
                changed = False
                if self.options.branch:
                    logger.info("Checking for changes since version %s; please "
                                "wait...", versions[-1])
                    changed = self.hasChangedSince(versions[-1],
                        self.options.branch)
                    if self.options.independent and not changed:
                        # only check if not already marked as changed
                        logger.info("Checking if last release is based on "
                                    "branch %s; please wait...",
                                    self.options.branch)
                        if not self.isLastReleaseFromBranch(versions[-1],
                            self.options.branch):
                            changed = True
                    if not changed:
                        logger.info("No changes detected.")
                else:
                    logger.info("Not checking for changes since version %s "
                                "because no -b or --use-branch was specified.",
                                versions[-1])
                # 3.3. If the branch changed and the next version should be
                # suggested, let's find the next version.
                if changed:
                    defaultVersion = base.guessNextVersion(versions[-1])
                else:
                    defaultVersion = versions[-1]
            else:
                logger.info("Not checking for changes because -n or "
                            "--next-version was not used")
                defaultVersion = versions[-1]
        else:
            logger.info(
                "Not checking for changes because --force-version was used")

        # If there's no version the package is probably non existent
        if defaultVersion is None and self.options.defaultPackageVersion:
            # avoid interactive questions (handy for automated builds)
            defaultVersion = self.options.defaultPackageVersion

        branch = self.options.branch
        while True:
            version = base.getInput(
                'Version for `%s`' %self.pkg, defaultVersion,
                self.options.useDefaults and defaultVersion is not None)
            if version not in versions and not self.options.offline:
                if askToCreateRelease:
                    print('The release %s-%s does not exist.' %(pkg, version))
                    doRelease = base.getInput(
                        'Do you want to create it? yes/no', 'yes',
                        self.options.useDefaults)
                    if doRelease == 'no':
                        continue
                # 4. Now create a release for this version.
                if not self.options.offline:
                    # 4.1. Determine the branch from which to base the release
                    # on.
                    if branch is None:
                        print('Available Branches:')
                        for branch in self.getBranches():
                            print('  * ' + branch)
                        print('  * trunk')
                        branch = base.getInput(
                            'What branch do you want to use?', 'trunk',
                            self.options.useDefaults)
                    # 4.2. Create the release.
                    self.createRelease(version, branch)
            break
        # 5. Return the version number.
        logger.info('Chosen version: ' + version)

        # save the info for build.py
        if branch is None:
            branch = 'trunk'
        self.branchUrl = self.getBranchURL(branch)
        self.branchRevision = self.getRevision(self.branchUrl)

        return version


def main(args=None):
    # Make sure we get the arguments.
    if args is None:
        args = sys.argv[1:]
    if not args:
        args = ['-h']

    # Set up logger handler
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(base.formatter)
    logger.addHandler(handler)

    # Parse arguments
    options, args = base.parser.parse_args(args)

    logger.setLevel(logging.INFO)
    if options.verbose:
        logger.setLevel(logging.DEBUG)
    if options.quiet:
        logger.setLevel(logging.FATAL)

    if len(args) == 0:
        print("No package was specified.")
        print("Usage: build-package [options] package1 package2 ...")
        sys.exit(0)
    for pkg in args:
        builder = PackageBuilder(pkg, options,
            forceSvnAuth=options.forceSvnAuth)
        try:
            builder.runCLI(options.configFile)
        except KeyboardInterrupt:
            logger.info("Quitting")
            sys.exit(0)

    # Remove the handler again.
    logger.removeHandler(handler)

    # Exit cleanly.
    sys.exit(0)



##############################################################################
#
# Copyright (c) 2015 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""

import os
from setuptools import setup, find_packages

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

setup (
    name='p01.build',
    version='0.8.5',
    author = "Roger Ineichen, Projekt01 GmbH",
    author_email = "dev@projekt01.ch",
    description = "A build system supporting application deploy and installation",
    long_description=(
        read('README.txt')
        + '\n\n' +
        read('src','p01','build','README.txt')
        + '\n\n' +
        read('CHANGES.txt')
        ),
    long_description_content_type = 'text/plain',
    license = "ZPL 2.1",
    keywords = "Zope3 z3c p01 build release deploy install",
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Framework :: Zope3'],
    url = 'https://pypi.org/pypi/p01.build',
    packages = find_packages('src'),
    include_package_data = True,
    package_dir = {'':'src'},
    namespace_packages = ['p01'],
    extras_require=dict(
        test=[
            'p01.checker',
            'zope.testing',
        ]),
    install_requires = [
        'beautifulsoup4',
        'setuptools',
        ],
    entry_points = {
        'console_scripts': [
            'build = p01.build.build:main',
            'build-package = p01.build.package:main',
            'deploy = p01.build.deploy:main',
            ],
        },
    zip_safe = False,
    )



##############################################################################
#
# Copyright (c) 2015 Projekt01 GmbH and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id:$
"""
__docformat__ = "reStructuredText"

import datetime
import bs4
import configparser
import io
import base64
import logging
import hashlib
import pkg_resources
import re
import sys
import shutil
import os
try:
    # py2
    from urllib2 import urlopen
    from urllib2 import Request
    from urllib2 import HTTPError
except ImportError:
    # py3
    from urllib.request import urlopen
    from urllib.request import Request
    from urllib.error import HTTPError

try:
    import pkg_resources._vendor.packaging.version
    packaging = pkg_resources._vendor.packaging
except ImportError:
    # fallback to naturally-installed version; allows system packagers to
    #  omit vendored packages.
    import packaging.version

from p01.build import base
from p01.build import package

logger = base.logger

is_win32 = sys.platform == 'win32'


def findProjectVersions(project, config, options, uploadType):
    if options.offline:
        logger.info('Offline: Skip looking for project versions.')
        return []

    VERSION = re.compile(project+r'-(\d+\.\d+(\.\d+){0,2})')

    if uploadType == 'local':
        dest = os.path.join(config.get(base.BUILD_SECTION, 'buildout-server'),
                            project)

        versions = []
        for root, dirs, files in os.walk(dest):
            for fname in files:
                m = VERSION.search(fname)
                if m:
                    versions.append(m.group(1))
    else:
        url = config.get(base.BUILD_SECTION, 'buildout-server') + project + '/'
        logger.debug('Package Index: ' + url)
        req = Request(url)

        username = config.get(base.BUILD_SECTION, 'buildout-server-username')
        password = config.get(base.BUILD_SECTION, 'buildout-server-password')
        credentials = ('%s:%s' % (username, password)).encode('utf-8')
        base64string = base64.b64encode(credentials).decode('utf-8')
        req.add_header("Authorization", "Basic %s" % base64string)

        try:
            soup = bs4.BeautifulSoup(urlopen(req).read(), "html.parser")
        except HTTPError as err:
            logger.error("There was an error accessing %s: %s" % (url, err))
            return []

        versions = []
        for tag in soup('a'):
            cntnt = str(tag.contents[0]) # str: re does not like non-strings
            m = VERSION.search(cntnt)
            if m:
                versions.append(m.group(1))

    return sorted(versions, key=lambda x: packaging.version.Version(x))


def getDependentConfigFiles(baseFolder, infile, addSelf=True, outfile=None,
                            hashes=None, options=None):
    # go and read all cfg files that are required by the master
    # to collect them all
    # if they have a path, modify according to that the the files are flat
    # on the server

    # baseFolder and infile might be "out of sync", because
    # we process cfg files that are already modified compared to the templates
    # in that case we want to read/write the modified file, but look for
    # the others in the template_path

    if hashes is not None:
        justname = os.path.split(infile)[-1]
        if justname not in hashes:
            hash = hashlib.md5(open(infile, 'rb').read()).hexdigest()
            hashes[justname] = hash

    config = base.NonDestructiveRawConfigParser()
    config.read(infile)

    dependents = set()
    if addSelf:
        dependents.add(infile)

    try:
        extends = config.get('buildout', 'extends')
    except configparser.NoSectionError:
        return dependents
    except configparser.NoOptionError:
        return dependents

    extendParts = extends.split()
    hasPath = False
    for part in extendParts:
        if '/' in part or '\\' in part:
            hasPath = True

        # extends filenames are always relative to the actual file
        fullname = os.path.join(baseFolder, part)

        if is_win32:
            #most buildouts use / but win32 uses \
            fullname = fullname.replace('/', '\\')

        if not os.path.exists(fullname):
            logger.error("FATAL: %s not found, but is referenced by %s" % (
                fullname, infile))
            sys.exit(0)

        dependents.update(getDependentConfigFiles(os.path.dirname(fullname),
            fullname, hashes=hashes, options=options))

    if hasPath:
        #we need to clean relative path from extends as on the server
        #everything is flat
        extendParts = [os.path.split(part)[-1] for part in extendParts]
        extends = '\n  '.join(extendParts)

        config.set('buildout', 'extends', extends)

        if outfile:
            #if the config is created by ourselves
            # config.write(open(outfile, 'w'))
            config.writeFile(outfile, options)
        else:
            #this is a referenced config, don't modify the original
            fName = os.path.split(infile)[-1]
            # config.write(open(fName, 'w'))
            config.writeFile(fName, options)

            if addSelf:
                #adjust dependents
                dependents.remove(infile)
                dependents.add(fName)

    return dependents


def addHashes(dependencies, hashes, rename=True, options=None):
    # add hashes to files

    rdep = []
    for fname in dependencies:
        modified = False
        justname = os.path.split(fname)[-1]

        config = base.NonDestructiveRawConfigParser()
        config.read(fname)

        try:
            # 1. modify file contents
            extends = config.get('buildout', 'extends')
            for oldname, hash in list(hashes.items()):
                parts = os.path.splitext(oldname)
                newname = "%s-%s%s" % (parts[0], hash, parts[1])

                if oldname in extends:
                    extends = extends.replace(oldname, newname)
                    modified = True

            if modified:
                config.set('buildout', 'extends', extends)
        except configparser.NoSectionError:
            pass
        except configparser.NoOptionError:
            pass

        # 2. rename/copy files
        newname = justname
        if rename:
            try:
                hash = hashes[justname]
                parts = os.path.splitext(justname)
                newname = "%s-%s%s" % (parts[0], hash, parts[1])
                modified = True
            except KeyError:
                pass

        if modified:
            #RawConfigParser fools us... next time it does I'll do
            #plain replace on the file contents
            # config.write(open(newname, 'w'))
            config.writeFile(newname, options)
            rdep.append(newname)
        else:
            rdep.append(fname)

    return rdep


def build(configFile, options):
    # save the time we started
    now = datetime.datetime.now()

    # Read the configuration file.
    logger.info('Loading configuration file: ' + configFile)
    config = base.NonDestructiveRawConfigParser()
    config.read(configFile)

    # Create the project config parser
    logger.info('Creating Project Configuration')
    projectParser = base.NonDestructiveRawConfigParser()
    template_path = None
    if config.has_option(base.BUILD_SECTION, 'template'):
        template = config.get(base.BUILD_SECTION, 'template')
        logger.info('Loading Project Configuration Template: ' + template)
        projectParser.read([template])
        template_path = os.path.abspath(template)

    if not projectParser.has_section('versions'):
        projectParser.add_section('versions')

    # Determine all versions of the important packages
    pkgversions = {}
    pkginfos = {}
    for pkg in config.get(base.BUILD_SECTION, 'packages').split():
        customPath = None
        if ':' in pkg:
            pkg, customPath = pkg.split(':')
        builder = package.PackageBuilder(pkg, options)

        version = builder.runCLI(configFile, askToCreateRelease=True,
                                 forceSvnAuth = options.forceSvnAuth)

        pkgversions[pkg] = version
        pkginfos[pkg] = (builder.branchUrl, builder.branchRevision)
        projectParser.set('versions', pkg, version)

    # Get upload type
    try:
        uploadType = config.get(base.BUILD_SECTION, 'buildout-upload-type')
    except configparser.NoOptionError:
        uploadType = "webdav"

    # Stop if no buildout-server given
    try:
        config.get(base.BUILD_SECTION, 'buildout-server')
    except configparser.NoOptionError:
        logger.info('No buildout-server specified in the cfg, STOPPING')
        logger.info('Selected package versions:\n%s' % (
            '\n'.join('%s = %s' % (pkg, version)
                      for pkg, version in list(pkgversions.items()))) )
        return

    # Write the new configuration file to disk
    projectName = config.get(base.BUILD_SECTION, 'name')
    defaultVersion = configVersion = config.get(base.BUILD_SECTION, 'version')
    projectVersions = findProjectVersions(projectName, config,
                                          options, uploadType)

    # Determine new project version
    if projectVersions:
        defaultVersion = projectVersions[-1]
    if options.nextVersion or configVersion == '+':
        defaultVersion = base.guessNextVersion(defaultVersion)
    if options.forceVersion:
        if options.forceVersion in projectVersions:
            logger.error('Forced version %s already exists' % \
                options.forceVersion)
        else:
            defaultVersion = options.forceVersion
    projectVersion = base.getInput(
        'Project Version', defaultVersion, options.useDefaults)

    # Write out the new project config -- the pinned versions
    simpleProjectFilename = '%s-%s.cfg' %(projectName, projectVersion)
    projectConfigFilename = base.getFilePath(simpleProjectFilename, options)
    logger.info('Writing project configuration file: ' + projectConfigFilename)
    # projectParser.write(open(projectConfigFilename, 'w'))
    projectParser.writeFile(projectConfigFilename, options)

    filesToUpload = [simpleProjectFilename]

    try:
        hashConfigFiles = config.getboolean(base.BUILD_SECTION,
                                            'hash-config-files')
    except configparser.NoOptionError:
        hashConfigFiles = False

    # Process config files, check for dependent config files
    # we should make sure that they are on the server
    # by design only the projectConfigFilename will have variable dependencies
    if template_path:
        if hashConfigFiles:
            hashes = {}
        else:
            hashes = None

        dependencies = getDependentConfigFiles(os.path.dirname(template_path),
            projectConfigFilename, addSelf=False, outfile=projectConfigFilename,
            hashes=hashes, options=options)

        if hashConfigFiles:
            dependencies = addHashes(dependencies, hashes, options=options)
            #fix main config too
            addHashes([projectConfigFilename], hashes, rename=False,
                options=options)

        filesToUpload.extend(dependencies)

    # Dump package repo infos
    # do it here, projectConfigFilename might be rewritten by
    # getDependentConfigFiles
    projectFile = open(projectConfigFilename, 'a')
    projectFile.write('\n')
    projectFile.write('# package SVN infos:\n')
    for pkg, pkginfo in list(pkginfos.items()):
        projectFile.writelines(
            ('# %s\n' % pkg,
             '#   svn URL:%s\n' % pkginfo[0],
             '#   svn repo revision:%s\n' % pkginfo[1][0],
             '#   svn last change revision:%s\n' % pkginfo[1][1],
            ))
        logger.info('SVN info: %s: %s %s %s', pkg, pkginfo[0],
                    pkginfo[1][0], pkginfo[1][1])
    projectFile.close()

    # Create deployment configurations
    # first define a parser for load addition variable (vars) section somewhere
    # in your templates chain.
    varsParser = base.NonDestructiveRawConfigParser()
    # make sure we use the right (tempalte inheritation) order
    varsParser.read(reversed(filesToUpload))

    for section in config.sections():
        if section == base.BUILD_SECTION:
            continue
        logger.info('Building deployment configuration: %s', section)
        template_path = config.get(section, 'template')
        logger.info('Loading deploy template file: %s', template_path)
        with open(template_path, 'r') as f:
            template = f.read()
        vars = dict([(name, value)
                     for name, value in config.items(section)
                     if name != 'template'])

        # apply additional vars defined in release template section if a
        # section name is defined as vars argument
        # or keep the vars argument as is if there is no section. This is
        # usefull if someone uses the vars attribute as argument (BBB)
        vName = vars.get('vars')
        if varsParser.has_section(vName):
            # apply vars from given section as additional vars
            for name, value in varsParser.items(vName):
                if name not in vars:
                    # but only if not overriden in deployment config
                    vars[name] = value

        # apply defaults
        vars['project-name'] = projectName
        vars['project-version'] = projectVersion
        vars['instance-name'] = section

        # add current time
        vars['current-datetime'] = now.isoformat()
        vars['current-date'] = now.date().isoformat()
        vars['current-time'] = now.time().isoformat()

        # not required anymore since we use input as is,
        # see: NonDestructiveRawConfigParser for more info
        ##handle multi-line items, ConfigParser removes leading spaces
        ##we need to add some back otherwise it will be a parsing error
        #for k, v in vars.items():
        #    if '\n' in v:
        #        #add a 2 space indent
        #        vars[k] = v.replace('\n', '\n  ')

        try:
            deployConfigText = template % vars
        except KeyError as e:
            logger.error(
                "The project variant %s template %s is missing the %r "
                "argument" % (section, template_path, e.message))
            sys.exit(0)
        except Exception as e:
            logger.error(
                "The project variant %s template %s probably provides unquoted "
                "attributes. Make sure all %% get prefixed ith an additonal %% "
                "(error: %r)" % (section, template_path, e.message))
            sys.exit(0)
        deployConfigFilename = '%s-%s-%s.cfg' %(
            config.get(base.BUILD_SECTION, 'name'), section, projectVersion)
        deployConfigFilename = base.getFilePath(deployConfigFilename, options)
        deployConfig = base.NonDestructiveRawConfigParser()
        deployConfig.readfp(io.StringIO(deployConfigText))
        deployConfig.set('buildout', 'extends', simpleProjectFilename)
        logger.info('Writing deployment file: ' + deployConfigFilename)
        # deployConfig.write(open(deployConfigFilename, 'w'))
        deployConfig.writeFile(deployConfigFilename, options)
        filesToUpload.append(deployConfigFilename)

    # Upload the files
    if uploadType == 'local':
        #no upload, just copy to destination
        dest = os.path.join(config.get(base.BUILD_SECTION, 'buildout-server'),
                            projectName)
        if not os.path.exists(dest):
            os.makedirs(dest)
        for filename in filesToUpload:
            fName = base.getFilePath(filename, options)
            justfname = os.path.split(fName)[-1]
            shutil.copyfile(fName, os.path.join(dest, justfname))
    elif uploadType == 'webdav':
        if not options.offline and not options.noUpload:
            for filename in filesToUpload:
                fName = base.getFilePath(filename, options)
                base.uploadFile(
                    fName,
                    config.get(
                        base.BUILD_SECTION, 'buildout-server')+'/'+projectName,
                    config.get(base.BUILD_SECTION, 'buildout-server-username'),
                    config.get(base.BUILD_SECTION, 'buildout-server-password'),
                    options.offline)
    elif uploadType == 'mypypi':
        if not options.offline and not options.noUpload:
            server = config.get(base.BUILD_SECTION, 'buildout-server')
            if not server.endswith('/'):
                server += '/'
            url = (server + projectName + '/upload')
            boundary = "--------------GHSKFJDLGDS7543FJKLFHRE75642756743254"
            headers={"Content-Type":
                "multipart/form-data; boundary=%s; charset=utf-8" % boundary}
            for filename in filesToUpload:
                fName = base.getFilePath(filename, options)
                justfname = os.path.split(fName)[-1]
                #being lazy here with the construction of the multipart form data
                content = """--%s
Content-Disposition: form-data; name="content";filename="%s"

%s
--%s--
""" % (boundary, justfname, open(fName, 'r').read(), boundary)

                base.uploadContent(
                    content, fName, url,
                    config.get(base.BUILD_SECTION, 'buildout-server-username'),
                    config.get(base.BUILD_SECTION, 'buildout-server-password'),
                    options.offline, method='POST', headers=headers)


def main(args=None):
    # Make sure we get the arguments.
    if args is None:
        args = sys.argv[1:]
    if not args:
        args = ['-h']

    # Set up logger handler
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(levelname)s - %(message)s')
    handler.setFormatter(base.formatter)
    logger.addHandler(handler)

    # Parse arguments
    options, args = base.parser.parse_args(args)

    logger.setLevel(logging.INFO)
    if options.verbose:
        logger.setLevel(logging.DEBUG)
    if options.quiet:
        logger.setLevel(logging.FATAL)

    try:
        build(options.configFile, options)
    except KeyboardInterrupt:
        logger.info("Quitting")
        sys.exit(0)

    # Remove the handler again.
    logger.removeHandler(handler)

    # Exit cleanly.
    sys.exit(0)
